/*
 * File:	aioli.h
 * Created:	8 July 2005
 * License:	GPL version 2
 * Authors:
 *		Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 * 		Adrien Lebre <adrien.lebre (at) imag.fr>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		Contains data structure declarations and function prototypes.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */

#ifndef AIOLI_H
#define AIOLI_H

#include "debug.h"

#include <linux/syscalls.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/hash.h>
#include <linux/vmalloc.h>

#include "iosched.h"		/* for private scheduler data ex. mlf_data */

#define PROC_STATS
#define LRU_CACHE
//#define LOG_FILE

#ifdef PROC_STATS
#ifndef LRU_CACHE
#error LRU_CACHE must be enabled when using PROC_STATS
#endif
#endif



#define HASH_SHIFT 6						/* 64 slots in */
#define HASH_ENTRIES		(1 << HASH_SHIFT)		/* hash table  */
#define HASH_FN(inode)		(hash_ptr(inode, HASH_SHIFT))

/*
 * Request type.
 */
enum {
	RT_READ = 0,
	RT_WRITE = 1,
};

/*
 * Request state.
 */
enum {
	RS_HASHTABLE = 0,
	RS_DISPATCH_QUEUE = 1,
	RS_PROCESSED = 2,
};

struct request_t;
/*
 * TODO: Following structure is my proposition for making aIOLi more generic
 */
struct client {

	/*
	 * Callback functions
	 */
	void (*process_request)(struct request_t *req);
	void (*process_requests)(struct list_head *reqs);

	/*
	 * aIOLi calls this to check if device holding file is idle.
	 * Should be supported by FS client
	 */
	int (*is_dev_idle)(void);
};

/*
 * We have one consumer thread running for each block device managed by
 * aIOLi, ex. if you have two hard drive then two consumer threads should
 * be running or if you have data stripping then you need one thread per
 * each device.
 * In general we call all those devices "io device".
 */
struct consumer_t {
	struct task_struct	*task;

	/* This is completed when thread exits. */
	struct completion	exited;

	/* This is completed when new request is added to dispatch_queue
	 * for a file that consumer is waiting for (deceptive idleness). */
	struct completion	request_added;
	
	/* IO scheduler used by this consumer. */
	struct io_scheduler_instance_t io_scheduler;

	/* TODO: This should be move to request_t because we have a consumer
	 * per "io device" and NOT a consumer per client. */
	struct client		*client;

	int io_device_id;

	struct list_head	list;
};

struct related_list_t {
	struct list_head list ;

	loff_t laststartoff ;	// Started offset for the last proceed request (by request 
									//	we mean 'virtual request'). 
	loff_t lastfinaloff ;	// As above but for final offset 
	loff_t predictedoff ;	//	When a request with a lower offset than the last is
									//	proceed, we compute a predicted offset for the next request
									//	During the next round, if the new request has an offset 
									//	greater than the predict one, a waiting time period is apply.
									//	This approach should fix inefficient shift phenomen.
									//	Please refer to aioli.imag.fr for more informations about 
									//	this problems.
	
	u32 lastaggregation ;	// Number of request contained in the last request.
	int proceedreq_nb;		// Number of read or write request proceed since the open 
									// of the file
	
	unsigned long nextquantum; // For QT purpose only
	
	struct request_file_t *req_file;

#ifdef PROC_STATS
	struct {
		u32	aggs_no;	 /* Number of aggregations proceed on
					  * on this file */
		u32	sum_of_agg_reqs; /* Sum of aggregated requests */
		u32	biggest;
		u32	shift_phenomens;
		u32	better_aggregation;
	} stats;
#endif
};

/*
 * This structure is used in hash table to store all request for particular
 * file (inode).
 */
struct request_file_t {
	void *file_id;
	struct related_list_t related_reads;
	struct related_list_t related_writes;
	struct list_head hashlist;
	struct list_head lru_list; /* To handle stateless filesystems */
};

typedef void internal_data;

struct io_data {
	loff_t offset;
	ssize_t len;
};

struct request_t {
	void *file_id;

	u64 timestamp;
	u64 jiffies_64;
	struct list_head timeline;

	/*
	 * These are use by hash table
	 */
	int type;
	struct list_head related;
	struct related_list_t *globalinfo;

	/*
	 * This is passed and used by process calling IO function.
	 */
	internal_data *data;

	/*
	 * Generic IO data used by aIOLi IO scheduler and hash table.
	 */
	struct io_data io_data;

	/*
	 * Request state, needed for I/O schedulers that track history
	 * of processed request.
	 */
	int state;
	
	u64 sched_factor;

#ifdef AIOLI_DEBUG
	int sanity;
#endif

	// TEMPORARY list to test responce threads!
	struct list_head list;
};

int __init aioli_init(struct client *clnt);
void __exit aioli_exit(void);

int aioli_add_request(void *file_id, int type, loff_t offset,
		       ssize_t len, internal_data *data, struct client *clnt);


// Try to add the request directly in the dispatch queue
int dispatch_queue_add(struct request_t *req, struct client *clnt);

/*
 * TODO: Fix cross reference.
 */
void process_requests(struct client *clnt, struct list_head *reqs);

/*
 * Statistics in /proc/aioli/stats
 */
#ifdef PROC_STATS
void stats_aggregation(struct related_list_t *related);
void specificstats_aggregation(struct related_list_t *related, u32 reqnb); // Adrien 
void stats_shift_phenomen(struct related_list_t *related);
void stats_better_aggregation(struct related_list_t *related);
#endif

#endif // #ifndef AIOLI_H

#ifdef LOG_FILE

#define LOG_PATH "/tmp/aioli.log"

int aioli_log_init(const char *log_file);
void aioli_log_proceed_request(struct request_t *req);
void aioli_log_proceed_requests(struct list_head *reqs);
void aioli_log_cleanup(void);

#endif

