/*
 * File:	consumer.c
 * Created:	26 July 2005
 * License:	GPL version 2
 * Authors:
 *    Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 *    Adrien Lebre <adrien.lebre (at) imag.fr>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		It parses the different lists stored in the hash table to find
 *		the right request to proccess.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */

#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/uaccess.h>
#include "aioli.h"
#include "request_cache.h"
#include "iosched.h"		// required for aIOLi scheduler (please see iosched.c)

int aioli_thread_stop = 0;

void process_requests(struct client *clnt, struct list_head *reqs)
{
	struct request_t *req;
	PRINT_FUNCTION_NAME;

#ifdef LOG_FILE
	aioli_log_proceed_requests(reqs);
#endif
	if (clnt->process_requests) {
		// WARNING : Currently we use naive approach which
		// implies to test if one request has been already processed
		// That means clnt->process_requests should do that ;)
		clnt->process_requests(reqs);
	} else {
		list_for_each_entry(req, reqs, related) {
			VERIFY_REQUEST(req);
			if (req->state == RS_DISPATCH_QUEUE) {
				clnt->process_request(req);
				req->state = RS_PROCESSED;
			}
		}
	}
}

static inline void
free_dispatch_queue(struct list_head *reqs)
{
	struct request_t *req;
	list_for_each_entry(req, reqs, related)
		request_cache_free(req);
}

static inline unsigned long consume_requests(struct consumer_t *consumer)
{
	struct io_scheduler_instance_t *sched;
	PRINT_FUNCTION_NAME;
	
	sched = &consumer->io_scheduler;

	if (sched->sched_ops->schedule(&sched->dispatch_queue, sched->sched_data))
		process_requests(consumer->client, &sched->dispatch_queue);

	return sched->sched_ops->post_process(&sched->dispatch_queue, sched->sched_data);
}

int
aioli_thread(void *arg)
{
	struct consumer_t *consumer = arg;
	unsigned long timeout=0;

	daemonize("aIOLi");
	consumer->task = current;
//	set_user_nice(current, -20);
	init_completion(&consumer->request_added);

	do {
	
	  // avoid 'active waiting' so :
	  // wait if a timeout is set or wait if there is no request
	   if (timeout > 0){
				wait_for_completion_timeout(&consumer->request_added,timeout);
		}
		else if(timeline_empty()){ 
			wait_for_completion(&consumer->request_added);
		}
	
	
		// This step can be reach if : 
		// timeout has been reached, new requests have been added
		// or aioli simply stop. So in this last case don't try to consume
		// some requests :p.
		if(!aioli_thread_stop){
			timeout = consume_requests(consumer);
			INIT_COMPLETION(consumer->request_added);
		}
	} while (!aioli_thread_stop);

	complete_and_exit(&consumer->exited, 0);
	consumer->task = 0;
	return 0;
}


