/*
 * File:	main.c
 * Created:	25 July 2005
 * License:	GPL version 2
 * Authors:
 *		Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 * 		Adrien Lebre <adrien.lebre (at) imag.fr>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		Contains code for module initialization and clean up.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/syscalls.h>
#include "aioli.h"
#include "iosched.h"
#include "request_cache.h"

void register_proc_entries(void);
void unregister_proc_entries(void);

extern struct task_struct *aioli_thread_task;
extern struct completion aioli_thread_exited;
int aioli_thread(void *);
extern int aioli_thread_stop;

/* This list keeps all consumer threads that are running */
static LIST_HEAD(consumers);

struct consumer_t consumer;

int start_consumer(struct client *clnt)
{
	int ret;

	init_completion(&consumer.exited);
	init_completion(&consumer.request_added);
	consumer.task = NULL;
	consumer.client = clnt;
	consumer.io_device_id = 0;

	ret = initialize_scheduler("quantum-mlf", &consumer.io_scheduler); 
//#warning Using MLF scheduler
//	ret = initialize_scheduler("mlf", &consumer.io_scheduler); 
	
	if (ret == 1) {
		xlist_add_tail(&consumer.list, &consumers);
		kernel_thread(aioli_thread, (void *)&consumer, CLONE_KERNEL);
	}

	return ret;
}

int __init aioli_init(struct client *clnt)
{
	int ret;
	PRINT_FUNCTION_NAME;

	if (!clnt || !clnt->process_request) {
		printk(KERN_ERR "Incorrect client structure\n");
		return -EINVAL;
	}


	register_static_io_schedulers();

	if ((ret = request_cache_init()))
		return ret;

	register_proc_entries();

#ifdef LOG_FILE
	ret = aioli_log_init(LOG_PATH);
	if (ret) {
		printk(KERN_ERR "Error opening log file %s: %d\n",
		       LOG_PATH, ret);
	}
#endif
	/*
	 * Now start consumer threads.
	 * TODO: Add configuration for those threads, I mean provide
	 * some way for user to tell aIOLi about "io nodes".
	 *
	 * TODO: Detect attaching of new devices (mounting a file system
	 * from a new device) and automatically start dedicated consumer.
	 */

	ret = start_consumer(clnt);
	if (ret != 1)
		aioli_exit();
	
	return 0;
}

void __exit aioli_exit(void)
{
	struct consumer_t *consumer;
	PRINT_FUNCTION_NAME;

	aioli_thread_stop = 1;

	list_for_each_entry(consumer, &consumers, list) {
		if (consumer->task) {
			complete(&consumer->request_added);
			wait_for_completion(&consumer->exited);
			if (consumer->io_scheduler.sched_ops->exit)
				consumer->io_scheduler.sched_ops->exit(consumer->io_scheduler.sched_data);
		}
	}

	unregister_proc_entries();

	request_cache_cleanup();
#ifdef LOG_FILE
	aioli_log_cleanup();
#endif
	printk(KERN_ALERT"aIOLi has been stopped\n");
}

//MODULE_LICENSE("GPL v2");
//MODULE_AUTHOR("Przemyslaw Sowa <sowa.przemyslaw (@) gmail.com>");
