/*
 * File:	iosched.c
 * Created:	2 August 2005
 * License:	GPL version 2
 * Authors:
 *    Adrien Lebre <adrien.lebre (at) imag.fr>
 *    Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 *
#include <linux/sched.h>
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		This file provides specific routines relating to the different
 *		scheduling strategies available in the aIOLi solution.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */
#include <linux/sched.h>
#include <linux/pagemap.h> 	// Use to compute LLONG_MAX value 
				// (PAGE_CACHE_SIZE in applyMLFonFile())

#include "aioli.h"
#include "iosched.h"
#include "request_cache.h"

/* **** ***************** **** *
 * Simple 'Time order' scheduler
 * **** ***************** **** */


// Only retrieve the oldest request presents in the aIOLi hashtable
// This request is added in the linked list 'req_list'.
loff_t simple_timeorder(struct list_head* req_list){
    struct request_t *req;		  
    req = timeline_oldest_req();
    if(req){
        hashtable_del_req(req);
        xlist_add(&req->related, req_list);
        return req->io_data.offset + req->io_data.len;
    }
    //else
    return 0;
}



/* **** ***************** **** *
 *  MLF schedule
 * **** ***************** **** */

#define MLF_QUANTUM 32768 // usual nfs rsize and wsize

/*
 * Note : HZ gives the number of clock cycle per second
 * so for some architecture where HZ equals 100, we can not wait less than 0.01 second
 */
#define WAIT_AGGREG_CONST ((HZ / 500) ? HZ / 500 : 1) // around 2ms  // was 6
#define WAIT_SHIFT_CONST ((HZ / 300) ? HZ / 300 : 1)   // around 3ms // was 10
									 				//(anticpatory uses 6 millisec)
													
//#define WAIT_AGGREG_CONST ((HZ / 150) ? HZ / 150 : 1) // around 6ms  
//#define WAIT_SHIFT_CONST ((HZ / 100) ? HZ / 100 : 1)   // around 10ms 
	

#define CHECK_AGGREGATE(req,nextreq) \
     ( (req->io_data.offset <= nextreq->io_data.offset)&& \
         ((req->io_data.offset+req->io_data.len)>=nextreq->io_data.offset))

// Returns the earliest timestamp for given list.
// This function is required to maintain coherency. 
static u64 find_timeposition(struct list_head* list)
{
        struct request_t *req;
    //	u64=(1<<(sizeof(u64)*8))-1;
    u64 stamp=~0;
    PRINT_FUNCTION_NAME;

    if(!list_empty(list)){
        list_for_each_entry(req,list,related){
            if(stamp <req->timestamp)
                stamp=req->timestamp;
        }
    }
    PRINT_FUNCTION_EXIT;
    return stamp;	
}

// reqlist : the related read or write linked list
// timeconstraint : the timestamp required to maintain coherency
// localMLF_l : reference  to  request selected by MLF algorithm.
// localstamp : used to select the MLF list between various files
// return the number of request present in the potential MLF request list
static inline u32 applyMLFonlist(struct list_head *reqlistheader,
                           u64 timeconstraint, struct list_head** localMLF_l, u64* localstamp)
{  
    // To handle the request_t cells
    struct list_head *req_l, *nextreq_l = NULL;
    struct request_t *req, *nextreq;
   
    // Beetween several requests processed by MLF 
    // algorithm for a same file, we select the 
    // smallest offset. By this way, we favour such
    // strategies such as 'read ahead'. 
    // Note : LLONG_MAX is not provided by kernel header. So we took a line 
	 // from include/linux/nfs_fh.h
    loff_t min_localoffset = (((loff_t) ULONG_MAX) << PAGE_CACHE_SHIFT) 
												+ PAGE_CACHE_SIZE - 1;
    
    u32 reqnb=0,tmpreqnb=0;
    
    // MLF required values for all 'virtual' request.
    struct list_head *tmpMLF_l;        
    u64 tmpstamp=0,tmpfactor=0,tmprealsize=0;			
    loff_t tmpoffset;  // Even if we favor smallest offset, we have to 
                       // take into account starvation problem.

	PRINT_FUNCTION_NAME; 
    // Now fetch the list
    req_l=reqlistheader->next;
    while(req_l != reqlistheader){

        req=list_entry(req_l, struct request_t, related);
        // Don't process request which doesn't fill the 
        // timeconstraint conditions (TODO check if we have
        // to increment their sched_factor since these 
        // requests are going to wait for one more cycle ??)
        while((req_l != reqlistheader)&&
                (req->timestamp > timeconstraint)){
            req_l=req_l->next;
                req=list_entry(req_l,struct request_t,related);
        }

	debug("parsed the list in progress");
	
        // We reach the end of the 'related_reads' list or
        // req references a 'coherency' compliant request 
        // that should process by MLF algorithm. 
        if(req_l != reqlistheader){
        	debug("one request could be compliant offset=%llu (minoff=%llu)"
				,req->io_data.offset,min_localoffset);

            // To favor 'read ahead' strategy per file
            // As the list is sorted by ascending order,
            // this test will pass just once !
	    // BUG : __u64 is required we don't know why  but in the other case
	    // the test will never pass ! Need time to fix it
            if((__u64)min_localoffset > (__u64)req->io_data.offset) {
                min_localoffset= req->io_data.offset;
		debug("min_localoffset was set to %llu", min_localoffset);
	} /* else {
		aioli_panic("min_localoffset = %llu\nio_data.offset = %llu",
			    min_localoffset, req->io_data.offset);
	} */

            // Start the 'virtual agregation' process
	    if (req->sched_factor==0) 
		req->sched_factor= 1; 
	    else
		req->sched_factor= req->sched_factor << 1; 
            
	    tmpfactor = req->sched_factor;
            tmpstamp = req->timestamp;
            tmpoffset = req->io_data.offset;
            tmprealsize =  req->io_data.len;
            tmpMLF_l=req_l;
            tmpreqnb=1;

	    debug("realsize %llu", tmprealsize);
            
	    nextreq_l=req_l->next;
            while ((nextreq_l != reqlistheader)&&
                     (nextreq=list_entry(nextreq_l, struct request_t,related))&&
                                         (CHECK_AGGREGATE(req,nextreq))){
                debug("starting aggregate process");

		    // Only aggregate request which doesn't impact on coherency!
                if(nextreq->timestamp <timeconstraint){
                    tmprealsize += nextreq->io_data.len;
                    tmpfactor+=nextreq->sched_factor;
                    tmpreqnb++;
                	debug("one more request added");
		
                    // fifo timestamp parameter used to select the right
                    // requests between to distinct file.
                    if(tmpstamp > nextreq->timestamp)
                        tmpstamp = nextreq->timestamp;
                     
                    req_l=nextreq_l;
                    req=nextreq;
                }
                nextreq_l=nextreq_l->next;
                // Notes, we've to handle peculiar case such as : 
                // timeconstraint=6 Read(0;10) t=1, Read(10;12) t=8, 
                // Read (10;20) t=2), where we have to bypass request 
                // Read(10;12) which could impacts on coherency but 
                // process the request Read (10;20). 

            } // end of while ((reqnext_l != reqlistheader)&&
              //                           (CHECK_AGREGATE(req,nextreq)))

           // This aggregation request is stopped here. We're going to start 
           // with the next aggregation request but before, we 've to check if
           // the current agregated request could be handle by the provided quantum.
           if(((tmpfactor*MLF_QUANTUM) >= tmprealsize) && 
                   (tmpoffset == min_localoffset)){ 
                    // TODO use a better junction coefficient to select the 
                    // smallest access.
                    // Indeed since the list is sorted by ascending order, this
                    // test will pass once and the smallest request will be
                    // always chosse.
               *localMLF_l=tmpMLF_l;
               *localstamp=tmpstamp;
	       reqnb=tmpreqnb;
               debug("one MLF local list has been selected");
	       debug("localMLF_l == NULL? - %d", (*localMLF_l == NULL));

           }
        } // end of if(req_l != reqlistheader) : 
          // if the next potential 'virtual' request has a timestamp greater 
          // than the provided timeconstraint

       req_l=nextreq_l;
    } // end of while(req_l != reqlistheader) :
      // Here, we've completely parsed the list.
      // So, whe have found a potential 'MLF requests' list or no ones could be
      // process with the provided quantum (maybe next round ;) !!)

    PRINT_FUNCTION_EXIT;
	       debug("localMLF_l == NULL? - %d", (*localMLF_l == NULL));
    return reqnb;
}

static inline u32 applyMLFonfile(struct request_file_t *reqfile, 
                           struct list_head **localMLF_l, u64 *localstamp)  
{
    u64 timeconstraint=0;			// To maintain coherency
    u32 reqnb=0;  
	PRINT_FUNCTION_NAME; 
    // First : on related read requests
    if (!list_empty(&reqfile->related_reads.list)){

        // To maintain coherency, we have to toake into account next write
        // timestamp request.
        // Can be improve in some kind of bytes ranges locking but need time ;)
		  // WARNING : Byte range locking requires fine management for example :
		  // One process create one file : the first byte is a flag to valid or 
		  // not the folowwing data (like a journal). In the first step, this flag 
		  // is set to , the the related data  are written and finally a last write 
		  // access is required to set the flag to 1. In this particular case, we 
		  // have to insure that the last write is well process before any read 
		  // even if they reference other offset.
        // Currently only read (and reciprocally write) requests with timestamp
        // lesser than the earliest write (and reciprocally read) could be 
        // selected.
        timeconstraint=find_timeposition(&reqfile->related_writes.list);
        reqnb=applyMLFonlist(&reqfile->related_reads.list,timeconstraint, localMLF_l, localstamp);
    }     
    // Since we don't use complex coherency mechanism such as bytes range
    // locking, if some request have been selected in the former part, we don't
    // have to test the related_writes list.
    // Note : better performance could provide by bytes rangel locking but ;)
    if(reqnb == 0){
       //Second : on related write requests
       if (!list_empty(&reqfile->related_writes.list)){
           timeconstraint=find_timeposition(&reqfile->related_reads.list);
          reqnb=applyMLFonlist(&reqfile->related_writes.list,timeconstraint, localMLF_l, localstamp);
       }
    }
    PRINT_FUNCTION_EXIT;
	       debug("localMLF_l == NULL? - %d", (*localMLF_l == NULL));
    return reqnb;
} 


// req_list : the list to build
// MLF_l the reference to the first request_t cells (shoudl not be NULL)
// returns the final offset, this is where read or write stops.
static inline loff_t buildMLFselectedlist(struct list_head *req_list, struct list_head *MLF_l)
{
    // To build the req_list;
    struct list_head *nextreq_l;
    struct request_t *req, *nextreq;
	 u32 potentialshift=0;			// If we detect a potential shift phenomen
	 										// we have to set predictedoff to finaloffset
    u64 timeconstraint=0;			// To maintain coherency
    struct related_list_t *related_list;
    
	PRINT_FUNCTION_NAME; 
	
    req = list_entry(MLF_l, struct request_t, related);
    related_list = req->globalinfo;
    if(req->type == RT_READ){
       timeconstraint=find_timeposition(&req->globalinfo->req_file->related_writes.list);
//	printk(KERN_ALERT"BUILD 1 READ: offset %llu length %ld\n",req->io_data.offset,req->io_data.len);
    }
    else{ // RT_WRITE
       timeconstraint=find_timeposition(&req->globalinfo->req_file->related_reads.list);
//	printk(KERN_ALERT"BUILD 1 WRITE: offset %llu length %ld\n",req->io_data.offset,req->io_data.len);
    }
   
	// Retrieve aggregation information for next schedule 
	 if ((req->io_data.offset<req->globalinfo->laststartoff)&&(!req->globalinfo->predictedoff)){ // Potential shift phenomen
		potentialshift=1;
	 }
	 
	 req->globalinfo->laststartoff=req->io_data.offset;
	 req->globalinfo->lastaggregation=1;
	 req->globalinfo->proceedreq_nb++;
	 req->globalinfo->lastfinaloff=req->io_data.offset + req->io_data.len;
    // Add the first request
    nextreq_l=MLF_l->next;
    timeline_del_req(req);
    __hashtable_del_req(req);
    req->state = RS_DISPATCH_QUEUE;
    xlist_add_tail(&req->related, req_list);
    
//relatedlist_display(&req->globalinfo->list);

   
    debug("Before loop");
    // Start the 'virtual agregation' process and build list in the meantime
    while ((nextreq_l != &related_list->list)&&
               (nextreq=list_entry(nextreq_l,struct request_t,related))&&
            (CHECK_AGGREGATE(req,nextreq))){
        
    debug("inside loop");
         nextreq_l=nextreq_l->next;
        // Only aggregate request which doesn't impact on coherency!
        if(nextreq->timestamp <timeconstraint){
            
            req=nextreq;
	 			req->globalinfo->lastaggregation++;
	 	req->globalinfo->proceedreq_nb++;
    debug("one more request ");
	    // Retrieve the real range for these accesses
	    if((req->io_data.offset+req->io_data.len)>req->globalinfo->lastfinaloff)
    	    	req->globalinfo->lastfinaloff=req->io_data.offset+req->io_data.len;
	    
            timeline_del_req(req);
            __hashtable_del_req(req);
    	    req->state = RS_DISPATCH_QUEUE;
            xlist_add_tail(&req->related, req_list);
//printk(KERN_ALERT"BUILD:  offset %llu length %ld\n",req->io_data.offset,req->io_data.len);
        }
        // Notes : we've to handle peculiar case such as : 
        // timeconstraint=6 Read(0;10) t=1, Read(10;12) t=8, 
        // Read (10;20) t=2), where we have to bypass request 
        // Read(10;12) which could impacts on coherency but 
        // process the request Read (10;20). 

    } // end of while ((reqnext_l != &related_list->list)&&
    //                           (CHECK_AGREGATE(req,nextreq)))

    debug("returning final_offset = %llu", related_list->lastfinaloff);

	 if (potentialshift){
	 // So set a predicted offset
		related_list->predictedoff= related_list->lastfinaloff;
	 }
    return related_list->lastfinaloff;
}   


static void aioli_wait(signed long timeout)
{
	PRINT_FUNCTION_NAME;
	//aioli_panic("entered, timeout = %l", timeout);
	
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(timeout);
	//yield();
}

// Return waiting time period in milli-second
// MLF_l references the 'virtual' request selected by the MLF algorithm
// reqnb is th enimber of request(s) inside this 'virtual' request
u32 checkSelection(struct list_head *MLF_l, u32 reqnb)
{
    struct request_t *req;
 	 PRINT_FUNCTION_NAME; 
	 req = list_entry(MLF_l, struct request_t, related);
	
	// No ckeck for the first request
	 if(req->globalinfo->proceedreq_nb==0){
	 	return 0;
	 }

	 // Check shift phenomen
	 // Please see aIOLi site for further informations
	 if(req->globalinfo->predictedoff!=0){ 
	 	// Set predicted to 0 since we gonna to 
		// wait for the estimated period and 
		// we don't want to imply starvation pb.
		req->globalinfo->predictedoff=0;

			if(req->io_data.offset>req->globalinfo->predictedoff){ 
			// if req->io_data.offset <= req->globalinfo->predictedoff the shift 
			// phenomem has been naturaly fixed
				print_red("shift phenomen detected");
#ifdef PROC_STATS
				stats_shift_phenomen(req->globalinfo);
#endif				
				return WAIT_SHIFT_CONST;
	 		}
	 }
	 // Check for better aggregation to prevent futur expensive seek operations
	 // TODO : pass the nb of rqst contained inside the selected virtual request.
	 if((req->io_data.offset>req->globalinfo->lastfinaloff)
						  && (req->globalinfo->lastaggregation > reqnb)){
			// As for shift phenomen, we've to  set to 0 to avoid starvation pb.
			req->globalinfo->lastaggregation=0;
			print_yellow("better aggregation could be done");
#ifdef PROC_STATS
				stats_better_aggregation(req->globalinfo);
#endif				
			return WAIT_AGGREG_CONST; 
	 }
	 return 0;
}

struct list_head* find_MLF(signed int *selectedindex, u32* globalreqnb)
{ 
    // To handle request_file_t objects
    int i;
    struct list_head* reqfile_l;
    struct request_file_t *req_file;
 	 
    // /LF references
    struct list_head *localMLF_l,*globalMLF_l;
    u64 localstamp,globalstamp=~0;
    u32 reqnb=0;
    u32 waitperiod=0;
	
    *globalreqnb=0;
	
    PRINT_FUNCTION_NAME;

	 globalMLF_l=NULL;
    // do ....
    // while MLF algorithm has not selected some requests whereas there are 
    // some.
    do {
        // Parse all requests to select the right one.
        // So for all hash entries
        for(i=0;i< hashtable_get_size();i++){

            reqfile_l=hashtable_lock(i);	
            // If there some related request_file_t cells
            // we have to parse them
            if(!list_empty(reqfile_l)){

                // For all file related to this hash, apply MLF algorithm
                list_for_each_entry(req_file, reqfile_l, hashlist){
                    
                    localMLF_l=NULL;
                    localstamp=~0;
                    reqnb=0;
                    
                    reqnb=applyMLFonfile(req_file, &localMLF_l,&localstamp);
                    if (reqnb>0){
								if(localstamp <= globalstamp){
   	                  // Some request have been selected by the MLF algorithm
         	            // So now apply the fifo approach between these requests
            	            globalstamp=localstamp;
               	         globalMLF_l=localMLF_l;
									*globalreqnb=reqnb;
		                     *selectedindex=i;
									debug("globalMLF selected");
                    	 }
						  } 
                	} // end list_for_each_entry(req_file, reqfile_l, hashlist)
            }
            hashtable_unlock(i)	;
		  } // end for(i=0;i<HASH_ENTRIES;i++)

		  if(globalMLF_l){ // If some requests have been selected, check if waiting 
								 // approach could lead to better performance
            
				reqfile_l=hashtable_lock(*selectedindex);	
				waitperiod=checkSelection(globalMLF_l,*globalreqnb);
            hashtable_unlock(*selectedindex)	;
				
				if(waitperiod){
					aioli_wait(waitperiod);
					debug("waitperiod = %u",waitperiod);
					globalMLF_l=NULL;
					*globalreqnb=0;
					globalstamp=~0;
					*selectedindex=-1;
				}
			} // End if checkSelection part
	 
	 }
    while((globalMLF_l == NULL)  && !timeline_empty() );
    // while MLF algorithm has not selected some requests whereas there are 
    // some.
	 
    return globalMLF_l;
}  

/* **** ***************** **** *
 * QT  schedule
 * **** ***************** **** */ 
// Please se WAIT_XXX macros for further informations
#define  ANTICIPATORY_VALUE ((HZ / 150) ? HZ / 150 : 1) // 6ms for HZ=1000 or 0.01s for UML architecture

int select_NMLF_reqs(struct list_head* reqs, unsigned long remaining_time, struct related_list_t* selectedlist)
{
    // To build the req_list;
    struct list_head *nextreq_l;
    struct request_t *req ;
    u64 timeconstraint=0;
	 u32 reqnb=0;
	 
 	 loff_t tmpfinaloff=0;
	  loff_t newfinaloff=selectedlist->lastfinaloff;
	  size_t tmplen;

	 PRINT_FUNCTION_NAME; 
	
	// Only requests that have startoffset >= to the laststartoff
	// and <= to the lastfinaloff could be added
	// coherency constraint shoud take into account too as well as 
	// remaining_time.
	
 
	 if(list_empty(&selectedlist->list)) {
		 //printk (KERN_ALERT"select_XXX: List is empty\n");
		return 0;
	}
//printk(KERN_ALERT"INSIDE SELECT\n");
//relatedlist_display(&selectedlist->list); 

	 // First : retrieve coherency constraint
	 nextreq_l=selectedlist->list.next;
    req = list_entry(nextreq_l, struct request_t, related);
    if(req->type == RT_READ){
       timeconstraint=find_timeposition(&req->globalinfo->req_file->related_writes.list);
    }
    else{ // RT_WRITE
       timeconstraint=find_timeposition(&req->globalinfo->req_file->related_reads.list);
    }
   
	// Second : Fetch the selectedlist to find compliant request 
	// and add it into the dispatch_queue.
	do{	
				
			//printk (KERN_ALERT"select_XXX: one more parse ;)\n");
			nextreq_l=nextreq_l->next;	

			if ((req->timestamp<timeconstraint) && // Coherency compliant
				 (req->io_data.offset >= selectedlist->laststartoff) // laststartoffset compliant
				&&(req->io_data.offset<=newfinaloff)) {// lastfinaloffset compliant 
				
				// This request might be insert into the dispatch queue
				// Last test : 'remaining time' compliant
				
				//printk (KERN_ALERT"select_XXX: one request is maybe compliant\n");
				if(req->io_data.offset+req->io_data.len<=newfinaloff){ 
					// All the data are already in the buffer cache : negligeable impact 
					// on remaining time. Just added it. 
    				timeline_del_req(req);
    				__hashtable_del_req(req); 	// Call __hashtable_del_req instead of 
														// hashtable_del_req since I already hold the lock.
					// TODO ADD IN THE DISPATCH QUEUE IN A SORTED ORDER
					req->state=RS_DISPATCH_QUEUE;
    				xlist_add_tail(&req->related, reqs);
// printk(KERN_ALERT"SELECT: offset %llu length %ld\n",req->io_data.offset,req->io_data.len);
				//printk (KERN_ALERT"select_XXX: add a new request already cached\n");
					reqnb++;
				
				}
				else{	// Compute the impact on remaining_time.
					tmpfinaloff=req->io_data.offset+req->io_data.len;
					tmplen=tmpfinaloff-newfinaloff; 
					// TODO find a way to verify that : 
					if(tmplen){ 	// if this lenght could be processed during 
							  	 		// the remaining time so added it.
						newfinaloff=tmpfinaloff;				
						// TODO ADD IN THE DISPATCH QUEUE IN A SORTED ORDER
						timeline_del_req(req);
    						__hashtable_del_req(req); 	// Call __hashtable_del_req instead of 
	
						req->state=RS_DISPATCH_QUEUE;
    					xlist_add_tail(&req->related, reqs);
//printk(KERN_ALERT"SELECT: offset %llu length %ld\n",req->io_data.offset,req->io_data.len);
				//printk (KERN_ALERT"select_XXX: add a new request not in the cache (new finaloff %llu last %llu)\n",newfinaloff,selectedlist->lastfinaloff);
						reqnb++;
					}
						
				}
			
			}
	}				 
	while ((nextreq_l != &selectedlist->list)&&
               (req=list_entry(nextreq_l,struct request_t,related)));
	if(reqnb){
		selectedlist->proceedreq_nb+=reqnb;
		selectedlist->lastaggregation+=reqnb;
		selectedlist->lastfinaloff=newfinaloff;
#ifdef PROC_STATS
		// TODO ONE BUG HERE
		// sometimes the sum number displayed in /proc/aioli/stat is greather than
		// the number of the request proceeded. Actually stats_aggregation add to 
		// the sum lastaggregation each time. For select_XXX_Reqs in should not 
		// but just add the number of request selected here.
		// To fix this bug as quickly (and Ugly ;)) as we can, we just add a 
		// temporary function : specificstat_aggregation which takes in argument 
		// this number of request
		// Update statistics info for /proc/aioli
 	 	specificstats_aggregation(selectedlist,reqnb);
#endif
	
	}
	return reqnb;
}	 

/* **** ***************** **** *
	   Exported functions 
 * **** ***************** **** */


/* 
 * GENERIC functions
 */
unsigned long generic_post_process(struct list_head *dispatch_queue, void *io_sched_data)
{
	struct request_t *req;
	PRINT_FUNCTION_NAME;

	list_for_each_entry(req, dispatch_queue, related) {
		VERIFY_REQUEST(req);
		request_cache_free(req);
	}
	print_yellow("dispatch queue count: %d", list_count(dispatch_queue));
	PRINT_FUNCTION_EXIT;
	return MAX_SCHEDULE_TIMEOUT;
}

/*
 * scheduler algorithms
 */

u32 timeorder(struct list_head* req_list, void *io_sched_data){
	INIT_LIST_HEAD(req_list);
    simple_timeorder(req_list);
	 return 1;
    //	aggregate_timeorder(req_list); // TODO
}

u32 WSJF(struct list_head* req_list, void *io_sched_data ){
    // TODO
	INIT_LIST_HEAD(req_list);
		 return 0;
}

u32 MLF(struct list_head* req_list, void *io_sched_data)
{	
	struct list_head* globalMLF_l=NULL;	
	signed int selectedindex=-1;
	u32 reqnb=0;

#ifdef PROC_STATS  	
	struct request_t *req;	
	struct related_list_t *globalinfo; 
#endif 

	INIT_LIST_HEAD(req_list);
	globalMLF_l=find_MLF(&selectedindex,&reqnb);

	if(globalMLF_l){
   // We found some request(s) so we build the related list

 		hashtable_lock(selectedindex);

#ifdef PROC_STATS  	
		// Update statistics info for /proc/aioli
    	req = list_entry(globalMLF_l, struct request_t, related);
		globalinfo = req->globalinfo;
#endif 
       buildMLFselectedlist(req_list, globalMLF_l);
       hashtable_unlock(selectedindex);

#ifdef PROC_STATS
		// Update statistics info for /proc/aioli
 	 	stats_aggregation(globalinfo);
#endif
		       
   }
	return reqnb;
}

#define QT_MAX_BOUND 500
#define FS_BANDWITH  (1024*1024*50) //~50MBytes // TODO find a way to determine dynamically this value

u32 QT(struct list_head* req_list, void *io_sched_data)
{
	unsigned long remaining_time;
	u32 reqnb=0;
	int hashindex=-1;
	struct request_t *req=NULL;
	
	struct qt_sched_data_t* qt_sched_data=(struct qt_sched_data_t*)io_sched_data;
	PRINT_FUNCTION_NAME;
	//printk (KERN_ALERT"QT inside\n");
	
	if(qt_sched_data->quantum == 0) {
		// If QT should select a new  'virtual' request
			INIT_LIST_HEAD(req_list);
  		   reqnb=qt_sched_data->base_sched.sched_ops->schedule(req_list,qt_sched_data->base_sched.sched_data);  
	
			// WARNING : below code are based on some specific fields of GLOBALINFO
			// GLOBALINFO is only set in MLF schedule. FIX theses following lines to 
			// provides QT with whatever scheduling strategies. TODO : The idea will consist in 
			// parsing the list to find the size and the other required informations
			if (reqnb){	
				// retrieve the related list:
				req = list_entry(req_list->next,struct request_t, related);
				qt_sched_data->related_list=req->globalinfo;	
				qt_sched_data->quantum=req->globalinfo->nextquantum;
				//printk (KERN_ALERT" QT quantum set to : %lu\n",qt_sched_data->quantum);
				qt_sched_data->start=jiffies;
			}
	}		
	else{
		// TODO look at problem 3 (see PROBLEMS file in the root directory)
  			  // Compute remaining time
		if ((jiffies - qt_sched_data->start) > qt_sched_data->quantum)
			remaining_time = 0;
		else {
			remaining_time = qt_sched_data->quantum - (jiffies - qt_sched_data->start);
			hashindex=HASH_FN(qt_sched_data->related_list->req_file->file_id);
			hashtable_lock(hashindex);
			reqnb=select_NMLF_reqs(req_list,remaining_time,qt_sched_data->related_list);
//			printk (KERN_ALERT"QT select reqs has been called and %d reqs have been selected\n",reqnb);
			//printk (KERN_ALERT"WARNING: select_reqs doesn't check remaining time TODO\n");
			hashtable_unlock(hashindex);
		}
	}
	PRINT_FUNCTION_EXIT;
	//printk (KERN_ALERT"QT outside\n");
	return reqnb;
}	

unsigned long QT_post_process(struct list_head* req_list, void *io_sched_data){

	struct qt_sched_data_t* qt_sched_data=(struct qt_sched_data_t*)io_sched_data;
	unsigned long timeout = 0;
	struct related_list_t *globalinfo=qt_sched_data->related_list;
	unsigned long requiredqt;

	PRINT_FUNCTION_NAME;
	//printk (KERN_ALERT"QT_post_process inside\n");

	//if timeout was reached : don't wait !!
	if ((jiffies - qt_sched_data->start) >= qt_sched_data->quantum){ 
		
		qt_sched_data->base_sched.sched_ops->post_process(req_list, qt_sched_data->base_sched.sched_data);
	
		// delete all the cell inside the dispatch queue;
		INIT_LIST_HEAD(req_list); // Ugly but efficient ;)

		//printk (KERN_ALERT"QT process quantum was : %lu\n",qt_sched_data->quantum);
		
		if(qt_sched_data->quantum==0){ 
			// First pass for this file : 
			// We have to compute its next quantum
			// We choose to assign twice the required quantum or in case at least ANTICIPATORY_VALUE
			// TODO find a way to pass dynamically FS bandwith
			requiredqt=(unsigned long)(globalinfo->lastfinaloff-globalinfo->laststartoff) * 2  * HZ / FS_BANDWITH ;
			//printk (KERN_ALERT"QT process next computed quantum : %lu (start: %lld  final: %lld)\n",requiredqt, globalinfo->laststartoff,globalinfo->lastfinaloff);
			globalinfo->nextquantum = (requiredqt > 0) ? globalinfo->nextquantum : ANTICIPATORY_VALUE;
		}
		else{
			requiredqt = (unsigned long)(qt_sched_data->related_list->lastfinaloff-qt_sched_data->related_list->laststartoff) * HZ / FS_BANDWITH;
			//printk (KERN_ALERT"QT process required quantum : %lu (start: %lld  final: %lld)\n",requiredqt, globalinfo->laststartoff,globalinfo->lastfinaloff);
		
			//Compute percentage of usage of the proposed qt
			if((requiredqt*100)/qt_sched_data->quantum > 80) 
					  // if it's bigger than 50% so give a better quantum
					qt_sched_data->related_list->nextquantum=qt_sched_data->quantum*2;
			else if((requiredqt*100)/qt_sched_data->quantum > 60) 
					qt_sched_data->related_list->nextquantum=qt_sched_data->quantum*15/10;
			else if((requiredqt*100)/qt_sched_data->quantum < 30) 
//					qt_sched_data->related_list->nextquantum=qt_sched_data->quantum*10/15;
					qt_sched_data->related_list->nextquantum = (requiredqt > 0) ? requiredqt : ANTICIPATORY_VALUE;
		
		  	if (qt_sched_data->related_list->nextquantum > QT_MAX_BOUND) 
				qt_sched_data->related_list->nextquantum=QT_MAX_BOUND;
		
			//Finally clean all qt_sched_data_t values: 
			qt_sched_data->quantum=0;
		}
		qt_sched_data->start=0;
		//printk (KERN_ALERT"QT process next processed quantum set to : %lu\n",	qt_sched_data->related_list->nextquantum);
	}
	else {
		//printk (KERN_ALERT"QT_post_process wait for one round more\n");
		timeout = qt_sched_data->quantum - (jiffies - qt_sched_data->start);
	}
	PRINT_FUNCTION_EXIT;
	//printk (KERN_ALERT"QT_post_process outside\n");
	return timeout;
}

// TODO please find a clean way to dynamically 
// call the sub init function.
// 
static int qt_mlf_init(void **io_sched_data)
{
//	struct qt_sched_data_t *data = (struct qt_sched_data_t *)io_sched_data;
	struct qt_sched_data_t *data=NULL;
	int ret = -ENOMEM;
	PRINT_FUNCTION_NAME;

	data = kmalloc(sizeof(struct qt_sched_data_t), GFP_KERNEL);
	if (data) {
		ret = initialize_scheduler("mlf", &data->base_sched);
		if (ret == 1) {
			data->related_list = NULL;
			data->quantum = 0;
			data->start = 0;
			*io_sched_data = data;
		} else
			kfree(data);
	}
	return ret;		
}

void qt_exit(void *io_sched_data)
{
	struct io_scheduler_instance_t *base_sched = 
		&((struct qt_sched_data_t *)io_sched_data)->base_sched;

	if (base_sched->sched_ops->exit)
		base_sched->sched_ops->exit(base_sched->sched_data);

	kfree(io_sched_data);
}



/**************************************
 *
 * I/O schedulers manager
 *
 **************************************/

// TODO comment ;)
// Return 1 if it succeeded
int initialize_scheduler(const char *name, struct io_scheduler_instance_t *instance) {
	int ret=1;
	PRINT_FUNCTION_NAME;
	
	instance->sched_ops = find_io_scheduler(name);
	
	if (instance->sched_ops == NULL)
		ret = -EINVAL;
	else if (instance->sched_ops->init)
			ret = instance->sched_ops->init(&instance->sched_data);

	return ret;
}

static LIST_HEAD(io_schedulers);

struct io_scheduler_ops *find_io_scheduler(const char *name)
{
	struct io_scheduler_ops *sched;
	PRINT_FUNCTION_NAME;
	
	list_for_each_entry(sched, &io_schedulers, list)
		if (strcmp(sched->name, name) == 0)
			return sched;
	return NULL;
}
		
int register_io_scheduler(struct io_scheduler_ops *io_sched)
{
	PRINT_FUNCTION_NAME;	

	if (!io_sched->name || !io_sched->schedule)
		return -1;

	list_add_tail(&io_sched->list, &io_schedulers);
	return 0;
}

void unregister_io_scheduler(const char *name)
{
	struct io_scheduler_ops *sched;
	PRINT_FUNCTION_NAME;

	sched = find_io_scheduler(name);
	if (sched)
		xlist_del(&sched->list);
}

void register_static_io_schedulers(void)
{
	static struct io_scheduler_ops scheds[] = {
		{
			.init = NULL,
			.exit = NULL,
			.schedule = &timeorder,
			.post_process = &generic_post_process,
			.name = "timeorder"
		},
		{
			.init = NULL,
			.exit = NULL,
			.schedule = &MLF,
			.post_process = &generic_post_process,
			.name = "mlf"
		},
		{
			.init = NULL,
			.exit = NULL,
			.schedule = &WSJF,
			.post_process = &generic_post_process,
			.name = "wsjf"
		},
		{
			.init = &qt_mlf_init,
			.exit = &qt_exit,
			.schedule = &QT,
			.post_process = &QT_post_process,
			.name = "quantum-mlf"
		}
	};
	int i = 0;
	PRINT_FUNCTION_NAME;

	for (i = 0; i < sizeof(scheds) / sizeof(scheds[0]); ++i)
		register_io_scheduler(&scheds[i]);
}

