/*
 * File:	iosched.h
 * Created:	3 August 2005
 * License:	GPL version 2
 * Authors:
 *    Adrien Lebre <adrien.lebre (at) imag.fr>
 *    Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		This file provides defines and prototypes relating to the different
 *		scheduling strategies available in the aIOLi solution.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */

#ifndef IOSCHED_H
#define IOSCHED_H


/* Defines */
#define TIME_SCHED 0
#define MLF_SCHED 1
#define WSJF_SCHED 2
#define NMLF_SCHED 3

#include <linux/list.h>

/*
 * Internal scheduler data, kept inside struct consumer_t for use by scheduler.
 */

struct nmlf_data {
	
	/* Quantum used when we're waiting for new request to arrive in
	 * hope to overcome deceptive idleness. */
	unsigned long		dispatch_quantum;

	/* If I/O scheduler is waiting for a request for a particular file
   	 * then this variable is set to file id. */	   
	void *			awaited_file_id;

	/* This is completed when new request for awaited file arrives. */
	struct completion	awaited_request_added;
};

/*
 *
 */
struct io_scheduler_instance_t {

	struct io_scheduler_ops		*sched_ops;

	/* Private I/O scheduler's data. */
	void				*sched_data;
	
	/* This queue contains all request that I/O scheduler decided they
	 * should be proceed now. */
	struct list_head 		dispatch_queue;
};


struct qt_sched_data_t {
	struct io_scheduler_instance_t	base_sched;

	struct related_list_t		*related_list;
	unsigned long			quantum;
	unsigned long			start;	
};

/*
 * Structure describing I/O scheduler operations.
 */
struct io_scheduler_ops {

	struct list_head list;
	
	/*
	 * Scheduler's name can be used to allow dynamically switch schedulers
	 * at runtime.
	 */
	const char *name;

	/*
	 * Called when new consumer thread using this scheduler is started.
	 * Intended to allow I/O scheduler to initialized it's private data.
	 */
	int (*init)(void **io_sched_data);	

	/*
	 * Called when scheduler is no longer use by some thread, but remember
	 * that it can be still used by other threads.
	 */
	void (*exit)(void *io_sched_data);
	
	/*
	 * This is the main scheduler function. It should add requests choosen
	 * to be processed now to @dispatch_queue.
	 *
	 * TODO: Find better name
	 * TODO: Find better return value (loff_t is ugly).
	 */
	u32 (*schedule)(struct list_head *dispatch_queue, void *io_sched_data);

	/*
   	 * Called by consumer when all request from dispatch queue are processed.
	 */	 
	unsigned long (*post_process)(struct list_head *dispatch_queue, void *io_sched_data);
};

/* Prototypes */
void register_static_io_schedulers(void);
struct io_scheduler_ops *find_io_scheduler(const char *name);
int initialize_scheduler(const char *name, struct io_scheduler_instance_t *instance);

/*
loff_t timeorder(struct list_head* req_list); 
loff_t MLF(struct list_head* req_list); 
loff_t WSJF(struct list_head* req_list); 
*/

#endif // #ifndef IOSCHED_H
