/*
 * File:	log.c
 * Created:	25 August 2005
 * License:	GPL version 2
 * Authors:
 *		Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 * 		Adrien Lebre <adrien.lebre (at) imag.fr>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		Contains code responsible for logging various (mainly)
 *		debuging information to a file.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */

#include <linux/module.h>
#include <asm/uaccess.h>	// for mm_segment
#include <linux/stat.h>		// for IS_I{R,W}USR
#include "aioli.h"

#ifdef LOG_FILE
//#define LOG_BUFFERED
#define LOG_PRINTK

/*
 * MAX(u32) 	= 4294967295		(10 characters)
 * MAX(loff_t) 	= 9223372036854775807	(19 characters)
 * MAX(ssize_t)	= 2147483647		(10 characters)
 */

unsigned long last_jiffies;

#ifndef LOG_PRINTK
#define LOG_BUFFER_SIZE 4096
static struct file *log_filp = NULL;
static char log_buffer[LOG_BUFFER_SIZE];
unsigned long last_jiffies;
#ifdef LOG_BUFFERED
static char *buf_ptr = log_buffer;
#endif
#endif /* #ifndef LOG_PRINTK */

/*
 * It there will be more the one consumer thread then we need to
 * protect log file with a lock.
 */
/* static DECLARE_MUTEX(logfile_lock); */

int aioli_log_init(const char *log_file)
{
#ifndef LOG_PRINTK
	char *tmp;
	int error, flags = O_WRONLY | O_CREAT | O_TRUNC; /* | O_NONBLOCK; */
	mm_segment_t oldfs;
	
	PRINT_FUNCTION_NAME;

#if BITS_PER_LONG != 32
	flags |= O_LARGEFILE;
#endif

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	tmp = getname(log_file);
	set_fs(oldfs);
	error = PTR_ERR(tmp);
	if (!IS_ERR(tmp)) {
		log_filp = filp_open(tmp, flags, S_IRUSR | S_IWUSR);
		error = PTR_ERR(log_filp);
		if (IS_ERR(log_filp))
			 log_filp = NULL;
		else {
			error = 0;
			last_jiffies = jiffies;
		}
		putname(tmp);
	}

	return error;
#else
	return 0;
#endif
}

#ifndef LOG_PRINTK
static inline loff_t file_pos_read(struct file *file)
{
	return file->f_pos;
}

static inline void file_pos_write(struct file *file, loff_t pos)
{
	file->f_pos = pos;
}

static inline void logfile_flush_buffer(void)
{
	mm_segment_t oldfs;
	loff_t pos;
	ssize_t ret, buf_len;
	
	PRINT_FUNCTION_NAME;
	debug("");

	buf_len = strlen(log_buffer);
	
	pos = file_pos_read(log_filp);
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	ret = vfs_write(log_filp, log_buffer, buf_len, &pos);
	set_fs(oldfs);
	file_pos_write(log_filp, pos);
#ifdef LOG_BUFFERED
	buf_ptr = log_buffer;
#endif
	if (ret != buf_len)
		print_red("vfs_write err: %d", ret);
		
//	log_buffer[0] = '\0';
}
#endif /* #ifndef LOG_PRINTK */

static inline void logfile_print_start(unsigned long file_id)
{
#ifndef LOG_PRINTK
#ifdef LOG_BUFFERED
	size_t free_space = LOG_BUFFER_SIZE - (int)(buf_ptr - log_buffer);
#endif
	PRINT_FUNCTION_NAME;
	/*
	 * It there will be more the one consumer thread then we need to
	 * protect log file with a lock.
	 */
	/* xdown(&logfile_lock); */
	
#ifdef LOG_BUFFERED
	if (free_space < 25) {
		logfile_flush_buffer();
		free_space = LOG_BUFFER_SIZE;
	}

	buf_ptr += snprintf(buf_ptr, free_space, "%07lu\t%lu: ", jiffies - last_jiffies, file_id);
#else
	snprintf(log_buffer, LOG_BUFFER_SIZE, "%07lu\t%lu: ", jiffies - last_jiffies, file_id);
	logfile_flush_buffer();		
#endif
#else /* #ifndef LOG_PRINTK */
	printk(KERN_ALERT "%07lu\t%lu: ", jiffies - last_jiffies, file_id);
#endif /* #ifndef LOG_PRINTK */
	last_jiffies = jiffies;
}

static void logfile_print_request(struct request_t *req)
{
#ifndef LOG_PRINTK
#ifdef LOG_BUFFERED
	size_t free_space = LOG_BUFFER_SIZE - (int)(buf_ptr - log_buffer);
#endif
	
	PRINT_FUNCTION_NAME;
	
#ifdef LOG_BUFFERED
	if (free_space < 45) {
		logfile_flush_buffer();
		free_space = LOG_BUFFER_SIZE;
	}

	buf_ptr += snprintf(buf_ptr, free_space, "(%lld; %lld) ",
			    req->io_data.offset,
			    req->io_data.offset + req->io_data.len);
#else
	snprintf(log_buffer, LOG_BUFFER_SIZE, "(%lld; %lld) ",
			    req->io_data.offset,
			    req->io_data.offset + req->io_data.len);
	logfile_flush_buffer();	
#endif 
#else /* #ifndef LOG_PRINTK */
	printk("(%lld; %lld) ",
			    req->io_data.offset,
			    req->io_data.offset + req->io_data.len);
#endif /* #ifndef LOG_PRINTK */
}

static inline void logfile_print_end(void)
{
#ifndef LOG_PRINTK
#ifdef LOG_BUFFERED
	size_t free_space = LOG_BUFFER_SIZE - (int)(buf_ptr - log_buffer);
#endif
	
	PRINT_FUNCTION_NAME;
	
#ifdef LOG_BUFFERED
	if (free_space < 2) {
		logfile_flush_buffer();
		free_space = LOG_BUFFER_SIZE;
	}

	buf_ptr += snprintf(buf_ptr, free_space, "\n");
#else
	snprintf(log_buffer, LOG_BUFFER_SIZE, "\n");
	logfile_flush_buffer();	
#endif
	
	/*
	 * It there will be more the one consumer thread then we need to
	 * protect log file with a lock.
	 */
	/* xup(&logfile_lock); */
	
#else /* #ifndef LOG_PRINTK */
	printk("\n");
#endif /* #ifndef LOG_PRINTK */
}

void aioli_log_proceed_request(struct request_t *req)
{
	PRINT_FUNCTION_NAME;

#ifndef LOG_PRINTK
	if (!log_filp)
		return;
#endif /* #ifndef LOG_PRINTK */
	
	logfile_print_start((unsigned long)req->file_id);
	logfile_print_request(req);
	logfile_print_end();
}

void aioli_log_proceed_requests(struct list_head *reqs)
{
	struct request_t *req;
	PRINT_FUNCTION_NAME;

#ifndef LOG_PRINTK
	if (!log_filp)
		return;
#endif /* #ifndef LOG_PRINTK */

	req = list_entry(reqs->next, struct request_t, related);
	logfile_print_start((unsigned long)req->file_id);

	list_for_each_entry(req, reqs, related)
		logfile_print_request(req);

	logfile_print_end();
}

void aioli_log_cleanup(void)
{
#ifndef LOG_PRINTK
	PRINT_FUNCTION_NAME;

	if (log_filp) {
#ifdef LOG_BUFFERED 
		if (buf_ptr != log_buffer)
			logfile_flush_buffer();
#endif
		fput(log_filp);
	}
#endif /* #ifndef LOG_PRINTK */
}

#endif /* #ifdef LOG_FILE */

