/*
 * File:	proc.c
 * Created:	26 July 2005
 * License:	GPL version 2
 * Authors:
 *		Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 * 		Adrien Lebre <adrien.lebre (at) imag.fr>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		Contains code for /proc virtual files implementation.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 */

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include "aioli.h"

struct proc_dir_entry *aioli_proc_dir;

#ifdef PROC_STATS
extern struct list_head lru_list;
extern atomic_t lru_count;

struct file_operations stats_proc_ops;
#endif

#ifdef AIOLI_DEBUG
extern struct file_operations hashtable_proc_ops;
#endif // #ifdef AIOLI_DEBUG

void register_proc_entries(void)
{
	struct proc_dir_entry *entry;

	PRINT_FUNCTION_NAME;

	aioli_proc_dir = proc_mkdir("aioli", NULL);
	if (!aioli_proc_dir) {
		printk(KERN_ERR "aIOLi: Cannot create /proc/aioli entry\n");  
		return;
	}

#ifdef PROC_STATS
	entry = create_proc_entry("stats", 444, aioli_proc_dir);
	if (entry) {
		entry->proc_fops = &stats_proc_ops;
	}
#endif 	/* #ifdef PROC_STATS */

#ifdef AIOLI_DEBUG
	entry = create_proc_entry("hashtable", 0, aioli_proc_dir);
	if (entry)
		entry->proc_fops = &hashtable_proc_ops;
#endif // #ifdef AIOLI_DEBUG
}

void unregister_proc_entries(void)
{

#ifdef PROC_STATS
	remove_proc_entry("stats", aioli_proc_dir);
#endif
	
#ifdef AIOLI_DEBUG
	remove_proc_entry("hashtable", aioli_proc_dir);
#endif // #ifdef AIOLI_DEBUG

	remove_proc_entry("aioli", NULL);
}

#ifdef PROC_STATS

void stats_aggregation(struct related_list_t *related)
{
	PRINT_FUNCTION_NAME;
	
	if (related->lastaggregation > 1) {
		related->stats.aggs_no++;
		related->stats.sum_of_agg_reqs += related->lastaggregation;
		if (related->stats.biggest < related->lastaggregation)
			related->stats.biggest = related->lastaggregation;
	}
	print_blue("last agg = %d, aggs = %d", related->lastaggregation,
		   related->stats.aggs_no);
	
}

// WARNING TEMPORARY FUNCTION REQUIRED BY SELECT_XXX_REQS INSIDE aioli/iosched.c
// Adrien
void specificstats_aggregation(struct related_list_t *related, u32 reqnb)
{
	PRINT_FUNCTION_NAME;
	
	// if there were just one request selected before that means we have 
	//a new aggregation request	
	if ((related->lastaggregation - reqnb) == 1) {
		related->stats.aggs_no++;
		related->stats.sum_of_agg_reqs += related->lastaggregation;
	}
	else 	// It was alread a virtual request so just add the number 
		// of new requests added
		related->stats.sum_of_agg_reqs += reqnb;

	if (related->stats.biggest < related->lastaggregation)
			related->stats.biggest = related->lastaggregation;

	print_blue("last agg = %d, aggs = %d", related->lastaggregation,
		   related->stats.aggs_no);
	
}



void stats_shift_phenomen(struct related_list_t *related)
{
	PRINT_FUNCTION_NAME;
	related->stats.shift_phenomens++;
}

void stats_better_aggregation(struct related_list_t *related)
{
	PRINT_FUNCTION_NAME;

	related->stats.better_aggregation++;
}

/*
 * /proc/aioli/stats
 */

static void *stats_seq_start(struct seq_file *s, loff_t *pos)
{
	PRINT_FUNCTION_NAME;

	if (list_empty(&lru_list))
		return NULL;
	
	if (*pos >= atomic_read(&lru_count)) 
		return NULL;
	if (*pos == 0) {
		seq_printf(s, "\t\t    \tavg\n");
		seq_printf(s, "\t\treqs\tsum / aggs\tbigg\tlast\tshift\tbetter\n");
	}

	return list_entry(lru_list.next, struct request_file_t, lru_list);
}

static void *stats_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	struct request_file_t *req_file;
	PRINT_FUNCTION_NAME;
	
	(*pos)++;
	if (*pos >= atomic_read(&lru_count)) 
		return NULL;
	
	req_file = (struct request_file_t *)v;
/*	
	if (req_file->lru_list.next == &lru_list) 
		return NULL;
*/
	return list_entry(req_file->lru_list.next, struct request_file_t,
			  lru_list);
}

static void stats_seq_stop(struct seq_file *s, void *v)
{
	PRINT_FUNCTION_NAME;
}

static int stats_seq_show(struct seq_file *s, void *v)
{
	struct request_file_t *req_file;
	struct related_list_t *related;
//TODO find a way to compute average in kernel mode
//	unsigned long average_aggs = 0;
	
	PRINT_FUNCTION_NAME;

	req_file = (struct request_file_t *)v;
	related = &req_file->related_reads;

// TODO compute in a clean way average value
//	if (related->stats.aggs_no > 0)
//		average_aggs = (unsigned long)(related->stats.sum_of_agg_reqs / 
//		   related->stats.aggs_no);
	
	seq_printf(s, "file id: %lx\n\tread:\t%u\t%u / %u\t%u\t%u\t%u\t%u\n",
		   (unsigned long)req_file->file_id,
		   related->proceedreq_nb,
		   related->stats.sum_of_agg_reqs,
		   related->stats.aggs_no,
		   related->stats.biggest,
		   related->lastaggregation,
		   related->stats.shift_phenomens,
		   related->stats.better_aggregation);


	related = &req_file->related_writes;

	seq_printf(s, "\twrite:\t%u\t%u / %u\t%u\t%u\t%u\t%u\n",
		   related->proceedreq_nb,
		   related->stats.sum_of_agg_reqs,
		   related->stats.aggs_no,
		   related->stats.biggest,
		   related->lastaggregation,
		   related->stats.shift_phenomens,
		   related->stats.better_aggregation);

	PRINT_FUNCTION_EXIT;
	return 0;
}

static ssize_t stats_proc_write(struct file * file, const char __user * buf,
				    size_t count, loff_t *ppos)
{
	struct request_file_t *req_file;
	PRINT_FUNCTION_NAME;

	list_for_each_entry(req_file, &lru_list, lru_list) {
		req_file->related_reads.laststartoff = 0;
		req_file->related_reads.predictedoff = 0;
		req_file->related_reads.lastaggregation = 0;
		req_file->related_reads.proceedreq_nb = 0;
		req_file->related_reads.stats.aggs_no = 0;
		req_file->related_reads.stats.sum_of_agg_reqs = 0;
		req_file->related_reads.stats.biggest = 0;
		req_file->related_reads.stats.shift_phenomens = 0;
		req_file->related_reads.stats.better_aggregation = 0;
		req_file->related_writes.laststartoff = 0;
		req_file->related_writes.predictedoff = 0;
		req_file->related_writes.lastaggregation = 0;
		req_file->related_writes.proceedreq_nb = 0;
		req_file->related_writes.stats.aggs_no = 0;
		req_file->related_writes.stats.sum_of_agg_reqs = 0;
		req_file->related_writes.stats.biggest = 0;
		req_file->related_writes.stats.shift_phenomens = 0;
		req_file->related_writes.stats.better_aggregation = 0;
	}
	return count;
}


static struct seq_operations stats_seq_ops = {
	.start		= stats_seq_start,
	.next		= stats_seq_next,
	.stop		= stats_seq_stop,
	.show		= stats_seq_show
};

static int stats_proc_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &stats_seq_ops);
}

struct file_operations stats_proc_ops = {
	.owner		= THIS_MODULE,
	.open		= stats_proc_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
	.write		= stats_proc_write,
};

#endif /* #ifdef PROC_STATS */

