/*
 * File:	request_cache.c
 * Created:	8 July 2005
 * License:	GPL version 2
 * Authors:
 *		Przemyslaw Sowa <sowa.przemyslaw (at) gmail.com>
 * 		Adrien Lebre <adrien.lebre (at) imag.fr>
 *
 * Description:
 *		This file is a part of the aIOLi kernel module implementation.
 *		Provides functions for managing request_t structures.
 *
 *		Further information is available at http://aioli.imag.fr
 *
 * Contributors:
 *		BULL Open Source Division
 *		INRIA France
 * 
 *		This program is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Changes:
 *	- 2 August 2005 - Przemyslaw Sowa
 *		Changed hash table implementation to use struct request_file_t
 *		as indirect data structure instead of keeping request_t
 *		directly on hash table list.
 */
 
#include <asm/semaphore.h>
#include <linux/hash.h>		/* for hash_ptr() */
#include <linux/bootmem.h>	/* for nr_kernel_pages */
#include <linux/seq_file.h>	/* for /proc/aioli_hashtable */
#include <linux/module.h>
#include <asm/atomic.h>
#include "aioli.h"

/*
 * TODO:
 * TMP TMP TMP TMP
 */
extern struct consumer_t consumer;

/*
 * Slab cache
 * It is used to manage frequently alocated structures:
 * request_t and request_file_t
 */
kmem_cache_t *request_cachep;
kmem_cache_t *request_file_cachep;

/*
 * Time line
 * It contains all requests, the same as hash table but sorted by arrival time.
 */
static LIST_HEAD(timeline);
static DECLARE_MUTEX(timeline_mutex);

/*
 * Hash table
 * Beacause we need to quickly find a place to store request for particular
 * file and/or find related requests, we use hash table that uses
 * request_t.file_id as value for hash function.
 */
static struct list_head *hashlist;
static struct semaphore *hashlist_locks;

static struct request_file_t *request_file_cache_alloc(void);
static void request_file_cache_free(struct request_file_t *req_file);

/*
 * LRU list of request_file_t
 * We use it to keep statistic information for file. Thanks to LRU list
 * we don't free request_file_t after all request for particular file
 * have been proceed, but we preserve this structure.
 * In case of having more then LRU_MAX_BOUND files inside hash table
 * we just take the oldest structure from LRU list at use it for new file.
 */
#ifdef LRU_CACHE

struct list_head lru_list;
atomic_t lru_count;
static DECLARE_MUTEX(lru_lock);
#define LRU_MAX_BOUND 255

#endif /* #ifdef LRU_CACHE */

//#ifdef AIOLI_DEBUG

/* Used only for debuging information */
int list_count(struct list_head *list)
{
	struct list_head *entry, *next = list->next;
	int count = 0;
	
	PRINT_FUNCTION_NAME;
	
	if (list_empty(list))
		return 0;
	
	printk(KERN_ALERT "list: %x -> ", (unsigned int)list);
	
	while ((entry = next) != list) {
		count++;
		next = entry->next;

		printk("%x -> ", (unsigned int)entry);
		if (count > 1000) {
			aioli_panic("Probably infinitive loop");
			return -1;
		}
	}
	
	return count;
}

/* Used only for debuging information */
int relatedlist_display(struct list_head *list)
{
	struct list_head *next = list->next;
	struct request_t *req;
	int count = 0;
	
	PRINT_FUNCTION_NAME;
	
	if (list_empty(list))
		return 0;
	
	
	while (next != list) {
		req = list_entry(next,struct request_t,related);
		printk(KERN_ALERT"\toffset %llu size %ld\n",req->io_data.offset,req->io_data.len);
		count++;
		next = next->next;
		if (count > 10000) {
			aioli_panic("Probably infinitive loop");
			return -1;
		}
	}
	
	PRINT_FUNCTION_EXIT;
	return count;
}
//#endif // #ifdef AIOLI_DEBUG

/*
 * Adds @req to the end of timeline list.
 *
 * Locking: 
 *	Must NOT have timeline_mutex
 */
void timeline_add_req(struct request_t *req)
{
	PRINT_FUNCTION_NAME;
	VERIFY_REQUEST(req);
	
	xdown(&timeline_mutex);
	
	if (list_empty(&timeline)) {
		req->timestamp = 0;
	} else {
		req->timestamp = ((struct request_t *)list_entry(timeline.prev,
							struct request_t,
							timeline))->timestamp + 1;
	}	
	req->jiffies_64 = get_jiffies_64();

	xlist_add_tail(&req->timeline, &timeline);

	xup(&timeline_mutex);
}

/*
 * Returns the oldest request and REMOVES it from timeline list.
 *
 * Returns:
 *	oldest request or NULL
 *
 * Locking: 
 *	Must NOT have timeline_mutex
 */
struct request_t *timeline_oldest_req(void)
{
	struct request_t *tmp;

	//PRINT_FUNCTION_NAME;

	xdown(&timeline_mutex);
	
	if (list_empty(&timeline)) {
		xup(&timeline_mutex);
		return 0;
	}

	tmp = list_entry(timeline.next, struct request_t, timeline);
	VERIFY_REQUEST(tmp);
	xlist_del(&tmp->timeline);

	xup(&timeline_mutex);
	return tmp;
}

/*
 * Returns the newest request and REMOVES it from timeline list.
 *
 * Returns:
 *	newest request or NULL
 *
 * Locking: 
 *	Must NOT have timeline_mutex
 */
struct request_t *timeline_newest_req(void)
{
	struct request_t *tmp;

	PRINT_FUNCTION_NAME;

	xdown(&timeline_mutex);
	
	if (list_empty(&timeline)) {
		xup(&timeline_mutex);
		return 0;
	}

	tmp = list_entry(timeline.prev, struct request_t, timeline);
	VERIFY_REQUEST(tmp);
	xlist_del(&tmp->timeline);

	xup(&timeline_mutex);
	return tmp;
}

/*
 * Removes @req from timeline list.
 *
 * Locking: 
 *	Must NOT have timeline_mutex
 */
void timeline_del_req(struct request_t *req)
{
	PRINT_FUNCTION_NAME;
	VERIFY_REQUEST(req);

	xdown(&timeline_mutex);

	xlist_del(&req->timeline);

	xup(&timeline_mutex);
}

/*
 * Informs if timeline list is empty
 *
 * Returns:
 *	1	- list is empty
 *	0	- otherwise
 */
inline int timeline_empty(void)
{
	return list_empty(&timeline);
}

/*
 * Function locks timeline.
 * Used by I/O scheduler before iterating the list.
 *
 * Locking:
 *	Must NOT be holding timeline_mutex.
 *	Must call timeline_unlock later.
 */ 
inline struct list_head *timeline_lock(void)
{
	down(&timeline_mutex);
	return &timeline;
}

/*
 * Function unlocks timeline.
 * Used by I/O scheduler after iterating through the list.
 *
 * Locking:
 *	Must be holding timeline_mutex.
 */ 
inline void timeline_unlock(void)
{
	up(&timeline_mutex);
}
/*
 * Allocates new struct request_t from slab cache.
 *
 * Returns:
 *	newly allocated struct request_t * or NULL
 */
struct request_t *request_cache_alloc(void)
{
	PRINT_FUNCTION_NAME;
	return kmem_cache_alloc(request_cachep, GFP_KERNEL);
}

/*
 * Frees @req by returning it to slab cache.
 */
void request_cache_free(struct request_t *req)
{
	PRINT_FUNCTION_NAME;
	VERIFY_REQUEST(req);
	kmem_cache_free(request_cachep, req);
}

static struct request_file_t *request_file_cache_alloc(void)
{
	PRINT_FUNCTION_NAME;
	return kmem_cache_alloc(request_file_cachep, GFP_KERNEL);
}

static void request_file_cache_free(struct request_file_t *req_file)
{
	PRINT_FUNCTION_NAME;
	kmem_cache_free(request_file_cachep, req_file);
}

/*
 * Function allocates new struct request_t from slab cache and sets it's fields
 * to values passed as arguments.
 */
struct request_t * request_constructor(void *file_id, int type, loff_t offset,
				       ssize_t len, internal_data *data) 
{
	struct request_t *new;

	PRINT_FUNCTION_NAME;

	new = request_cache_alloc();
	if (!new)
		return new;

	new->file_id = file_id;
	new->type = type;
	new->data = data;
	new->io_data.offset = offset;
	new->io_data.len = len;
	new->sched_factor = 0;
	new->state = RS_HASHTABLE;

#ifdef AIOLI_DEBUG
	debug("type = %d\noffset = %lu\nlen = %lu", new->type,
	      (unsigned long)new->io_data.offset, (unsigned long)new->io_data.len);
	new->sanity = 123456;
#endif
	return new;
}

/*
 * Function sets fields of struct request_file_t @req_file to default values.
 */
static void request_file_init(struct request_file_t *req_file, void *file_id)
{
	req_file->file_id = file_id;

	INIT_LIST_HEAD(&req_file->related_reads.list);
	req_file->related_reads.laststartoff = 0;
	req_file->related_reads.lastfinaloff = 0;
	req_file->related_reads.predictedoff = 0;
	req_file->related_reads.lastaggregation = 0;
	req_file->related_reads.proceedreq_nb = 0;
	req_file->related_reads.nextquantum = 0;
	req_file->related_reads.req_file = req_file;

	INIT_LIST_HEAD(&req_file->related_writes.list);
	req_file->related_writes.laststartoff = 0;
	req_file->related_writes.lastfinaloff = 0;
	req_file->related_writes.predictedoff = 0;
	req_file->related_writes.lastaggregation = 0;
	req_file->related_writes.proceedreq_nb = 0;
	req_file->related_writes.nextquantum = 0;
	req_file->related_writes.req_file = req_file;

#ifdef PROC_STATS

	/*
	 * memset part was commented out, because it didn't
	 * work as expected, but maybe it should be tested again.
	 */
/*
	memset(&req_file->related_reads.stats, 0,
	       sizeof(req_file->related_reads.stats));
	memset(&req_file->related_writes.stats, 0,
	       sizeof(req_file->related_writes.stats));
*/

	req_file->related_reads.stats.aggs_no = 0;
	req_file->related_reads.stats.sum_of_agg_reqs = 0;
	req_file->related_reads.stats.biggest = 0;
	req_file->related_reads.stats.shift_phenomens = 0;
	req_file->related_reads.stats.better_aggregation = 0;
	
	req_file->related_writes.stats.aggs_no = 0;
	req_file->related_writes.stats.sum_of_agg_reqs = 0;
	req_file->related_writes.stats.biggest = 0;
	req_file->related_writes.stats.shift_phenomens = 0;
	req_file->related_writes.stats.better_aggregation = 0;

#endif
}

/*
 * Function allocates new struct request_file_t from slab cache
 * and initializes it's fields to default values.
 *
 * Arguments:
 *	@file_id	- id of the file we're allocating structure for
 *
 * Returns:
 *	struct request_file * or NULL
 */
struct request_file_t * request_file_constructor(void *file_id)
{
	struct request_file_t *req_file;

	PRINT_FUNCTION_NAME;

	req_file = request_file_cache_alloc();
	if (!req_file)
		return req_file;

	request_file_init(req_file, file_id);
	return req_file;
}

#ifdef LRU_CACHE
/*
 * Function takes the last frequently used struct request_file_t from lru_list
 * and initializes it's fields to default values.
 * It isn't done for performance (we're using slab cache for it) but to keep
 * statistic information about a file as long as possible - until we have
 * LRU_MAX_BOUND structures and we started to overwrite the last freq. used.
 *
 * Arguments:
 *	@file_id	- id of the file we're allocating structure for
 *
 * Returns:
 *	struct request_file * / must succeed
 *
 * Locking:
 *	Must NOT be holding lru_lock.
 */
struct request_file_t * request_file_constructor_from_lru(void *file_id)
{
	struct request_file_t *req_file;
	PRINT_FUNCTION_NAME;
	
	xdown(&lru_lock);
	req_file = list_entry(lru_list.prev, struct request_file_t, lru_list);

	/*
	 * Remove it from the list, so it won't be taken by someone else.
	 * Also initialized list head, to tell update_lru_list() that we've
	 * just taken this entry for a new file.
	 */
	list_del_init(&req_file->lru_list);
	xup(&lru_lock);
	
	request_file_init(req_file, file_id);
	return req_file;		
}
#endif /* #ifdef LRU_CACHE */

/*
 * Function called whenever a struct request_t is freed to check
 * if we have other request for this file, if not then we free
 * related struct request_file_t.
 * In case of using LRU list for request_file_t we don't do anything.
 *
 * Arguments:
 *	@req_file	- the one we want to check for being used
 *
 * Locking:
 *	Must be holding hashlist_locks[hash_val]
 */
static inline void request_file_put(struct request_file_t *req_file)
{
	if (list_empty(&req_file->related_reads.list) && 
	    list_empty(&req_file->related_writes.list)) {
#ifndef LRU_CACHE
		xlist_del(&req_file->hashlist);
		request_file_cache_free(req_file);
#endif
	}
}

/*
 * This function creates all SLAB caches that we use and 
 * allocates/initializes related locks (semaphores).
 */
int request_cache_init(void)
{
	int i;

	PRINT_FUNCTION_NAME;
	
#ifdef LRU_CACHE
	INIT_LIST_HEAD(&lru_list);
	atomic_set(&lru_count, 0);
#endif /* #ifdef LRU_CACHE */

	INIT_LIST_HEAD(&timeline);
	request_cachep = kmem_cache_create("request", sizeof(struct request_t),
					   0, 0 /* maybe SLAB_HWCACHE_ALIGN? */,
					   NULL, NULL);
	if (!request_cachep) {
		printk(KERN_ERR "aIOLi: cannot create request SLAB cache\n");
		return -ENOMEM;
	}
	
	hashlist = kmalloc(sizeof(struct list_head) * HASH_ENTRIES, GFP_KERNEL);
	if (!hashlist) {
		printk(KERN_ERR "aIOLi: cannot allocate memory for req cache\n");
		return -ENOMEM;
	}

	request_file_cachep = kmem_cache_create("request_file",
						 sizeof(struct request_file_t),
					   	 0, 0, NULL, NULL);
	if (!request_file_cachep) {
		printk(KERN_ERR "aIOLi: cannot create request_file SLAB cache\n");
		return -ENOMEM;
	}
	
	hashlist_locks = kmalloc(sizeof(struct semaphore) * HASH_ENTRIES,
				 GFP_KERNEL);
	if (!hashlist_locks) {
		printk(KERN_ERR "aIOLi: cannot allocate memory for req locks\n");
		kfree(hashlist);
		return -ENOMEM;
	}
	
	for (i = 0; i < HASH_ENTRIES; i++) {
		INIT_LIST_HEAD(&hashlist[i]);
		init_MUTEX(&hashlist_locks[i]);
	}
	
	return 0;
}

/*
 * Function destroys all SLAB caches previously created by request_cache_init().
 */
void request_cache_cleanup(void)
{
#ifdef LRU_CACHE
	struct request_file_t *req_file;
#endif

	PRINT_FUNCTION_NAME;

	if (kmem_cache_destroy(request_cachep))
		printk(KERN_ERR "aIOLi: cannot free request SLAB cache\n");

#ifdef LRU_CACHE
	/*
	 * When LRU cache for request_file_t is used, then we don't
	 * free those structures, so we need to do this now.
	 */
	list_for_each_entry(req_file, &lru_list, lru_list)
		request_file_cache_free(req_file);
#endif
	
	if (kmem_cache_destroy(request_file_cachep))
		printk(KERN_ERR "aIOLi: cannot free request_file SLAB cache\n");

	kfree(hashlist);
	kfree(hashlist_locks);
}

#ifdef LRU_CACHE

/*
 * This function is called whenever new struct request_t is added to hash table.
 * It is here to reflect usage for particular struct request_file_t by moving
 * this request_file_t to the begining or LRU list.
 */
static void update_lru_list(struct request_file_t *req_file)
{
	PRINT_FUNCTION_NAME;
	

	if (list_empty(&req_file->lru_list)) {
		/*
		 * If list is empty it means that this is a newly allocated
		 * structure (or taken from LRU list and re-initialized)
		 * and we don't need to remove it from LRU list first.
		 */
		xdown(&lru_lock);
		xlist_add(&req_file->lru_list, &lru_list);
		xup(&lru_lock);
	} else {
		if (lru_list.next != &req_file->lru_list) {
			/*
			 * If this element isn't the first already then
			 * move it to the begining.
			 */
			xdown(&lru_lock);
			xlist_del(&req_file->lru_list);
			xlist_add(&req_file->lru_list, &lru_list);
			xup(&lru_lock);
		}
	}
}

#endif /* #ifdef LRU_CACHE */

/*
 * This is internal function which adds @req to hash table and places it
 * according to @hash_val.
 *
 * Locking:
 * 	Must be holding hashlist_locks[hash_val]
 */
static void __hashtable_add_req(struct request_t *req, unsigned long hash_val)
{
	struct list_head *hash_list = &hashlist[hash_val];
	struct request_file_t *req_file, *tmp_file = NULL;
	struct request_t *tmp;

	PRINT_FUNCTION_NAME;

	debug("list count: %d", list_count(hash_list));

#ifdef AIOLI_DEBUG
	/*
	 * This is a debuging code that prints list for this slot and
	 * checks if it is well sorted.
	 */
	//printk(KERN_ALERT "hash_list: ");
	list_for_each_entry(req_file, hash_list, hashlist) {
		//	printk("%lu -> ", (unsigned long)req_file->file_id);
		if (tmp_file)
			if (tmp_file->file_id > req_file->file_id)
				aioli_panic("Bad sorted list");
		tmp_file = req_file;
	}
	//printk("\n");
#endif

	/*
	 * Iterate through sorted collison list for this slot searching for
	 * entry about the file that @req treats with.
	 */
	list_for_each_entry(req_file, hash_list, hashlist) {
		if (req_file->file_id >= req->file_id) {
			break;
		}
	}
	
	/*
	 * If we don't have other request for this file, we need to allocate new
	 * structure request_file_t.
	 */
	if (req_file->file_id != req->file_id) {
#ifdef LRU_CACHE
		/*
		 * If we haven't exceed LRU_MAX_BOUND then allocate new
		 * structure from slab cache.
		 */
	  	if (atomic_read(&lru_count) < LRU_MAX_BOUND) {
			print_red("Allocating new request");
			req_file = request_file_constructor(req->file_id);
			atomic_inc(&lru_count);
			INIT_LIST_HEAD(&req_file->lru_list);
		} 
		else {
			/*
			 * We have many request_file_t structures already, so
			 * take the last frequently used from LRU list.
			 */
			print_green("Taking request from LRU list");
			req_file = request_file_constructor_from_lru(req->file_id);
		}
#else
		/*
		 * If we don't use LRU list to preserve information about previous
		 * accesses to a file then just allocate new structure from
		 * slab cache.
		 */
		req_file = request_file_constructor(req->file_id);

#endif /* #ifdef LRU_CACHE */

		/*
		 * Add to collision list in sorted order
		 */
		list_for_each_entry(tmp_file, hash_list, hashlist)
			if (tmp_file->file_id > req_file->file_id)
				break;

		xlist_add(&req_file->hashlist, tmp_file->hashlist.prev);
	}

#ifdef LRU_CACHE
	/*
	 * Update usage information for request_file_t related to
	 * newly added @req.
	 */
	update_lru_list(req_file);
	print_blue("LRU count: %d", list_count(&lru_list));
#endif /* #ifdef LRU_CACHE */

	/*
	 * Now we have request_file_t structure describing our file,
	 * so add @req to one of it's lists.
	 */
	
	/* Choose appropriate list. */
	if (req->type == RT_READ) {
		hash_list = &req_file->related_reads.list;
		req->globalinfo = &req_file->related_reads;
	} else {
		hash_list = &req_file->related_writes.list;
		req->globalinfo = &req_file->related_writes;
	}

	/* Search for proper position for @req (the list is sorted). */
	if (!list_empty(hash_list)) {
		list_for_each_entry(tmp, hash_list, related) {
			if (tmp->io_data.offset > req->io_data.offset ||
			    (tmp->io_data.offset == req->io_data.offset &&
			    tmp->io_data.len > req->io_data.len)) {
				hash_list = &tmp->related;
				break;
			}
		}
	}
	
	/* Add @req to the list in sorted order. */
	xlist_add(&req->related, hash_list->prev);

	/* Add @req to timeline list */
	timeline_add_req(req);

	/*
	 * TODO: Add support for many consumer threads.
	 */
	
	
	/* Signalize to the consumer thread that new request was added. */
	complete(&consumer.request_added);
	PRINT_FUNCTION_EXIT;
}

/*
 * Function adds @req to hash table.
 *
 * Arguments:
 *	@req	- request to be added.
 *	@clnt	- structure describing calling client.
 *
 * Locking:
 *	Must NOT hold hastlist_locks[hash_val]
 */
void hashtable_add_req(struct request_t *req, struct client *clnt)
{
	unsigned long hash_val;

	PRINT_FUNCTION_NAME;
	VERIFY_REQUEST(req);
	
	
	hash_val = HASH_FN(req->file_id);	
	down(&hashlist_locks[hash_val]);
	
	__hashtable_add_req(req, hash_val);

	xup(&hashlist_locks[hash_val]);
}

/* Internal function (but called by I/O scheduler)
 * that removes @req from hash table.
 *
 * Locking:
 *	Must be holding hashlist_locks[hash_val]
 */
void __hashtable_del_req(struct request_t *req)
{
	xlist_del(&req->related);
	request_file_put(req->globalinfo->req_file);
//	req->globalinfo = NULL;  // Temporary commented to test the qt approach
}

/*
 * Function removes @req from hash table.
 *
 * Locking:
 *	Must NOT be holding hashlist_locks[hash_val]
 */ 
void hashtable_del_req(struct request_t *req)
{
	unsigned long hash_val = HASH_FN(req->file_id);

	PRINT_FUNCTION_NAME;
	VERIFY_REQUEST(req);

	down(&hashlist_locks[hash_val]);
	
	__hashtable_del_req(req);

	xup(&hashlist_locks[hash_val]);
}

/*
 * Function locks hash table slot (collision list) identified by @index.
 * Used by I/O scheduler before iterating through collision list.
 *
 * Locking:
 *	Must NOT be holding hashlist_locks[hash_val]
 *	Must call hashtable_unlock(@index) later.
 */ 
struct list_head *hashtable_lock(int index)
{
	down(&hashlist_locks[index]);
	return &hashlist[index];
}

/*
 * Function unlocks hash table slot (collision list) identified by @index.
 * Used by I/O scheduler after iterating through collision list.
 *
 * Locking:
 *	Must be holding hashlist_locks[hash_val]
 */ 
void hashtable_unlock(int index)
{
	up(&hashlist_locks[index]);
}

int aioli_add_request(void *file_id, int type, loff_t offset,
		       ssize_t len, internal_data *data, struct client *clnt)
{
	struct request_t *req;
	
	PRINT_FUNCTION_NAME;
	
	debug("rqstp addr = %x", (unsigned int)data);

	req = request_constructor(file_id, type, offset, len, data);
	if (req) {
		hashtable_add_req(req, clnt);	
		return 0;		
	}		
	return -ENOMEM;
}

int hashtable_get_size(void)
{
	return HASH_ENTRIES;
}

#ifdef AIOLI_DEBUG

/*
 * /proc interface for dumping hash table
 */

static void *hashtable_seq_start(struct seq_file *s, loff_t *pos)
{
	if (*pos >= HASH_ENTRIES) 
		return NULL;

	return pos; 
}

static void *hashtable_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	(*pos)++;
	
	if (*pos >= HASH_ENTRIES) 
		return NULL;

	return pos;
}

static void hashtable_seq_stop(struct seq_file *s, void *v)
{
	PRINT_FUNCTION_NAME;
}

static int hashtable_seq_show(struct seq_file *s, void *v)
{
	int index = *(int *)v;
	struct list_head *hash_list = &hashlist[index];
	struct request_file_t *req_file;
	struct request_t *tmp;

	PRINT_FUNCTION_NAME;

	xdown(&hashlist_locks[index]);

	seq_printf(s, "[%d]\n", index);
	if (list_empty(hash_list))
		goto out;

	list_for_each_entry(req_file, hash_list, hashlist) {
		seq_printf(s, "     [%x]:", (unsigned int)req_file->file_id);
		seq_printf(s, "\n\t reads : ");
		list_for_each_entry(tmp, &req_file->related_reads.list,
				    related) {
			seq_printf(s, "(%d; %d), ", (int)tmp->io_data.offset,
		       	       (int)tmp->io_data.offset +
			       (int)tmp->io_data.len);
		}
		
		seq_printf(s, "\n\t writes: ");
		list_for_each_entry(tmp, &req_file->related_writes.list,
				    related) {
			seq_printf(s, "(%d; %d), ", (int)tmp->io_data.offset,
			       (int)tmp->io_data.offset +
			       (int)tmp->io_data.len);
		}
		seq_printf(s,"\n");
	}
out:
	xup(&hashlist_locks[index]);
	PRINT_FUNCTION_EXIT;
	return 0;
}
static struct seq_operations hashtable_seq_ops = {
	.start		= hashtable_seq_start,
	.next		= hashtable_seq_next,
	.stop		= hashtable_seq_stop,
	.show		= hashtable_seq_show
};

static int hashtable_proc_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &hashtable_seq_ops);
}

struct file_operations hashtable_proc_ops = {
	.owner		= THIS_MODULE,
	.open		= hashtable_proc_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

#endif // #ifdef AIOLI_DEBUG

